
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "m/m_def.h"

#include "m/m_random.h"

/* Some C++ compilers need this. (not g++). */
#ifdef __cplusplus
extern "C" {
#endif

/* -------------------------------------------------------------------------- */

#define GAUSSIAN_NOISE_LUT_SIZE   10000
#define ROOT_ONE_HALF_PIE         1.253314137
#define GAUSSIAN_NOISE_LUT_JUMP   (ROOT_ONE_HALF_PIE / GAUSSIAN_NOISE_LUT_SIZE)
#define GAUSSIAN_NOISE_LUT_RESOLUTION 100.0

/* -------------------------------------------------------------------------- */

static double* fs_gaussian_distribution_lut = NULL;

/* Will be at least one less than the table size. */
static int   fs_gaussian_distribution_lut_size;

/* -------------------------------------------------------------------------- */

static int initialize_gaussian_distribution_lut(void);

#ifdef TRACK_MEMORY_ALLOCATION
    static void prepare_memory_cleanup(void);
    static void free_gaussian_distribution_lut(void);
#endif

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                        get_gauss_random_matrix
 *
 * Gets a Gaussian random matrix
 *
 * This routine gets a matrix of the specified dimensions, and fills it with
 * gaussian random values with mean 0.0 and variance 1.0. The routine kjb_rand()
 * is used.
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is.
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails.
 *
 * Index: matrices, random
 *
 * -----------------------------------------------------------------------------
*/

int get_gauss_random_matrix(Matrix** mpp, int num_rows, int num_cols)
{
    int i, j;

    ERE(get_target_matrix(mpp, num_rows, num_cols));

    for (i = 0; i < num_rows; i++)
    {
        for (j = 0; j < num_cols; j++)
        {
            (*mpp)->elements[ i ][ j ] = gauss_rand();
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                        get_gauss_random_matrix_2
 *
 * Gets a Gaussian random matrix
 *
 * This routine is exaclty like get_gauss_random_matrix(), exept that the
 * alternative random stream (e.g. kjb_rand_2()) is used.
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails.
 *
 * Index: matrices, random
 *
 * -----------------------------------------------------------------------------
*/

int get_gauss_random_matrix_2(Matrix** mpp, int num_rows, int num_cols)
{
    int i, j;

    ERE(get_target_matrix(mpp, num_rows, num_cols));

    for (i = 0; i < num_rows; i++)
    {
        for (j = 0; j < num_cols; j++)
        {
            (*mpp)->elements[ i ][ j ] = gauss_rand_2();
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                        get_gauss_random_vector
 *
 * Gets a Gaussian random vector
 *
 * This routine gets a matrix of the specified length, and fills it with
 * Gaussian random values with mean 0.0 and variance 1.0. The routine kjb_rand()
 * is used.
 *
 * The first argument is the adress of the target vector. If the target vector
 * itself is null, then a vector  of the appropriate size is created. If the
 * target vector is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is.
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails.
 *
 * Index: vectors, random
 *
 * -----------------------------------------------------------------------------
*/

int get_gauss_random_vector(Vector** vp, int length)
{
    int i;

    ERE(get_target_vector(vp, length));

    for (i = 0; i < length; i++)
    {
        (*vp)->elements[ i ] = gauss_rand();
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                        get_gauss_random_vector_2
 *
 * Gets a Gaussian random vector
 *
 * This routine is exaclty like get_gauss_random_vector(), exept that the
 * alternative random stream (e.g. kjb_rand_2()) is used.
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails.
 *
 * Index: vectors, random
 *
 * -----------------------------------------------------------------------------
*/

int get_gauss_random_vector_2(Vector** vp, int length)
{
    int i;

    ERE(get_target_vector(vp, length));

    for (i = 0; i < length; i++)
    {
        (*vp)->elements[ i ] = gauss_rand_2();
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * Alternative, non-closed form way, that uses lookup_gauss_rand(). Not
 * recomended, but used for sanity checking. It is also faster than using
 * gauss_rand(), but the loss of purity is not likely warranted.
*/
int get_lookup_gauss_random_vector(Vector** vp, int length)
{
    int i;

    ERE(get_target_vector(vp, length));

    for (i = 0; i < length; i++)
    {
        (*vp)->elements[ i ] = lookup_gauss_rand();
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                               gauss_rand
 *
 * Returns a Gaussian distributed, random number
 *
 * This routine returns an approximately  Gaussian distributed, random number,
 * with mean zero and variance one. The random stream from kjb_rand() is used.
 *
 * Index: random
 *
 * See also:
 *     kjb_rand
 *
 * -----------------------------------------------------------------------------
*/

/*
 * Use the Box-Muller method. We used the random stream from kjb_rand(). Gaurd
 * against floating point exceptions.
*/
double gauss_rand()
{
  double theta = M_PI * kjb_rand();
  /*
   *  Get exponantially distributed (lamda=0.5) r_squared
   *  (-1/p)*log(uniform) is distributed as exp(); here p is 1/2.
  */
  double r_squared = -2.0 * SAFE_LOG(kjb_rand());

  if (r_squared <= 0.0)
  {
      return 0.0;
  }
  else
  {
      return sqrt(r_squared) * cos(theta);
  }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                               gauss_rand_2
 *
 * Returns a Gaussian distributed, random number
 *
 * This routine returns an approximately  Gaussian distributed, random number,
 * with mean zero and variance one. The random stream from kjb_rand_2() is used.
 *
 * This routine is exactly the same as gauss_rand(), except that the alternative
 * random stream kjb_rand_2() is used.
 *
 * Index: random
 *
 * See also:
 *     kjb_rand
 *
 * -----------------------------------------------------------------------------
*/

/*
 * Use the Box-Muller method. We used the random stream from kjb_rand(). Gaurd
 * against floating point exceptions.
*/
double gauss_rand_2()
{
  double theta = M_PI * kjb_rand_2();
  /*
   *  Get exponantially distributed (lamda=0.5) r_squared
   *  (-1/p)*log(uniform) is distributed as exp(); here p is 1/2.
  */
  double r_squared = -2.0 * SAFE_LOG(kjb_rand_2());

  if (r_squared <= 0.0)
  {
      return 0.0;
  }
  else
  {
      return sqrt(r_squared) * cos(theta);
  }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * I wrote this version before I was aware that there was closed form solution.
 * Keep this for now, mainly for debugging. It is approximately twice as fast as
 * gauss_rand(), but at the expense of quality of the random numbers. It could
 * be useful where extreme speed is necessary, or as a template for sampling a
 * distribution where there is no closed form.
*/
double lookup_gauss_rand(void)
{
    double uniform_random_number;
    double gaussian_random_number;
    int    negative_noise         = FALSE;
    double a, x, y;
    int    index;

    EPETE(initialize_gaussian_distribution_lut());

    /*
    // We use kjb_rand, as opposed to kjb_rand_2, to make this routine part of
    // the first random stream.
    */
    uniform_random_number = kjb_rand();

    if (uniform_random_number >= 0.5)
    {
        negative_noise = TRUE;
        uniform_random_number -= 0.5;
    }

    index = (int)(uniform_random_number * 2.0 * fs_gaussian_distribution_lut_size);

    if ((index < 0) || (index >= fs_gaussian_distribution_lut_size))
    {
        SET_CANT_HAPPEN_BUG();
        kjb_exit(EXIT_FAILURE);
    }

    ASSERT(fs_gaussian_distribution_lut_size < GAUSSIAN_NOISE_LUT_SIZE);

    x = fs_gaussian_distribution_lut[ index ];
    y = fs_gaussian_distribution_lut[ index + 1];
    /*
    // We use kjb_rand, as opposed to kjb_rand_2, to make this routine part of
    // the first random stream.
    */
    a = kjb_rand();

    gaussian_random_number = a * x + (1.0 - a) * y;

    if (negative_noise)
    {
        gaussian_random_number *= -1.0;
    }

    return gaussian_random_number;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int initialize_gaussian_distribution_lut(void)
{
    long_double    sum      = 0.0;
    int            count    = 0;
    double         x        = 0.0;
    double         step     = GAUSSIAN_NOISE_LUT_JUMP /
                                                  GAUSSIAN_NOISE_LUT_RESOLUTION;
    double         dist_val = 1.0;
    int            num_its  = 0;

    if (fs_gaussian_distribution_lut != NULL) return NO_ERROR;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    NRE(fs_gaussian_distribution_lut = DBL_MALLOC(GAUSSIAN_NOISE_LUT_SIZE));

    fs_gaussian_distribution_lut[ 0 ] = 0.0;

    /*CONSTCOND*/
    while (TRUE)
    {
        if ((num_its > 10000) || (count >= GAUSSIAN_NOISE_LUT_SIZE - 1))
        {
            break;
        }

        if (sum > GAUSSIAN_NOISE_LUT_JUMP)
        {
            sum -= GAUSSIAN_NOISE_LUT_JUMP;
            count++;

            fs_gaussian_distribution_lut[ count ] = x;
            step = GAUSSIAN_NOISE_LUT_JUMP / dist_val /
                                                  GAUSSIAN_NOISE_LUT_RESOLUTION;
            num_its = 0;
        }

        x += step;
        dist_val = exp(-x*x/2.0);
        sum += dist_val * step;

        num_its ++;
    }

    fs_gaussian_distribution_lut_size = count;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE;

    if (first_time)
    {
        add_cleanup_function(free_gaussian_distribution_lut);
        first_time = FALSE;
    }
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_gaussian_distribution_lut(void)
{
    kjb_free(fs_gaussian_distribution_lut);
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef __cplusplus
}
#endif

